(* this unit GUI.pas are part of Koganion IDE (RqWork6)
   copyright (c)2020 vasile eodor nastasa
   web: http://www.rqwork.de
   mail: nastasa.eodor@gmail.com
*)

unit GUI;

interface

uses
  Windows, Messages, SysUtils,Variants, {} Classes, {}Graphics, TypInfo;

type
   QStringList=class(TStringList);
   QFont=class(TFont);
   QApplication=class;
   QObject=class;
   QComponent=class;
   QControl=class;
   QWinControl=class;
   QForm=class;

   QEvent=procedure(Sender :QObject) of object;
   QCloseEvent=procedure(Sender :QObject;var CloseAction :integer) of object;
   QMessageEvent=procedure(var m:tagMSG) of object;
   QCommandEvent=procedure(Sender :QObject;iD, Code:integer; Dlg:hwnd) of object;

   QMessage=record
      Dlg, msg, wparam, lparam, result :integer;
      Sender, Captured :QControl;
   end;

   QObject=class(TPersistent)
   public
     Name, ClassName, ClassAncestor :string;
     constructor Create(AOwner :QComponent); dynamic; abstract;
   end;

   QComponent=class(QObject)
   private
      fOwner:QComponent;
      fComponents :QStringList;
   public
      constructor Create(AOwner :QComponent); override;
      destructor Destroy; override;
      property Components:QStringList read fComponents;
   end;

   QControl=class(QComponent)
   private
      fLeft,fTop,fWidth,fHeight :integer;
      fText:string;
      fEnabled, fVisible:boolean;
      fParent:QWinControl;
      fControls :QStringList;
      procedure SetParent(v :QWinControl);
      procedure SetText(v:string);
      procedure SetEnabled(v:boolean);
      procedure SetVisible(v:boolean);
      procedure SetLeft(v:integer);
      procedure SetTop(v:integer);
      procedure SetWidth(v:integer);
      procedure SetHeight(v:integer);
   protected
      fHandle:hwnd;
      procedure CreateHandle; virtual; abstract;
      procedure DoMessage(var m:QMessage); virtual; abstract;
      procedure Handler(var m:QMessage); virtual; abstract;
      //function RegisterClass:integer; virtual; abstract;
   public
      constructor Create(AOwner :QComponent);override;
      destructor Destroy; override;
      property Controls:QStringList read fControls;
   published
      property Parent: QWinControl read fParent write SetParent;
      property Text :string read fText write SetText;
      property Left :integer read fLeft write SetLeft;
      property Top :integer read fLeft write SetTop;
      property Width :integer read fLeft write SetWidth;
      property Height :integer read fLeft write SetHeight;
      property Enabled:boolean read fEnabled write SetEnabled;
      property Visible:boolean read fVisible write SetVisible;
   end;

   QWinControl=class(QControl)
   private
      fStyle, fExStyle  :cardinal;
      fGrouped, fTabStop :boolean;
      fParentWindow:hwnd;
      fFont:QFont;
      procedure SetFont(v:QFont);
      procedure SetStyle(v:cardinal);
      procedure SetExStyle(v:cardinal);
      procedure SetParentWindow(v:hwnd);
      procedure SetGrouped(v:boolean);
      procedure SetTabStop(v:boolean);
   protected
      procedure CreateHandle; override;
      procedure Handler(var m:QMessage); override;
      procedure DoMessage(var m:QMessage); override;
      
   public
      constructor Create(AOwner :QComponent);override;
      destructor Destroy; override;
      function RegisterClass:integer; //override;
   published
      property Font:QFont read fFont write SetFont;
      property Handle:hwnd read fHandle;
      property ParentWindow:hwnd read fParentWindow write SetParentWindow;
      property Style:cardinal read fStyle write SetStyle;
      property ExStyle:cardinal read fStyle write SetExStyle;
      property Grouped:boolean read fGrouped write SetGrouped;
      property TabStop:boolean read fTabStop write SetTabStop;
   end;

   QWinControls = class of QWinControl;

   QForm=class(QWinControl)
   private
      fCreate :QEvent;
      fClose :QCloseEvent;
      fCommand :QCommandEvent;
   protected
      procedure CreateHandle; override;
      procedure DoMessage(var m:QMessage); override;
      procedure Handler(var m:QMessage); override;
   public
      constructor Create(AOwner :QComponent);override;
      destructor Destroy; override;
   published
      property OnCreate :QEvent read fCreate write fCreate;
      property OnClose :QCloseEvent read fClose write fClose;
      property OnCommand :QCommandEvent read fCommand write fCommand;
   end;

   QApplication=class(QComponent)
   private
     fForms :QStringList;
   protected
      procedure Process(var m:tagMSG); stdcall;
   public
      constructor Create; reintroduce;
      destructor destroy ; override;
      procedure DoEvents;
      procedure Run;
      procedure Terminate;
      procedure Quit;
      procedure CreateForm(Ref :QWinControls;var F :Pointer);
      property Forms:QStringList read fForms;
   end;

var
   App:QApplication;
   wc :tagWNDCLASSEX;
   CreateControl :QWinControl;
   GlobalMsg:tagMSG;

const
    CM_COMMAND = WM_APP+100;
    CM_NOTIFY = WM_APP+101;
    CM_PARENTNOTIFY = WM_APP+102;


implementation

uses
   StandardCtrls;

function isClassExists(v:string):boolean;
begin
    wc.cbSize:=sizeof(wc);
    result :=(GetClassInfoEx(0,PChar(v),wc) or GetClassInfoEx(hinstance,PChar(v),wc));
end;

function GlobalDlgProc(dlg,msg,wparam,lparam:integer):lresult;stdcall;
var
   Ctrl:QWinControl;
   m:QMessage;
   cn:string;
   c:integer;
   cbf:array[0..254] of char;
begin
    m.msg:=msg;
    m.Dlg:=dlg;
    m.wparam:=wparam;
    m.lparam:=lparam;
    m.result:=0;
    m.Sender:=CreateControl;
    c:=GetClassName(dlg,cbf,255);
    cn:=cbf;
    cn:=trim(copy(cn,1,c));
    if m.sender<>nil then
       m.sender.ClassName:=cn;

    if CreateControl<>nil then begin
       Ctrl:=CreateControl;
       Ctrl.ClassName:=cn;
       if msg=wm_nccreate then begin
          Ctrl.fHandle:=dlg;
          SetWindowLong(dlg,GetClassLong(dlg,gcl_cbwndextra)-4,longint(Ctrl));
          CreateControl:=nil;
       end;
    end else
       Ctrl:=QWinControl(GetWindowLong(dlg,GetClassLong(dlg,gcl_cbwndextra)-4));

    if Ctrl<>nil then begin
       Ctrl.DoMessage(m);
       if m.result>=1 then begin
          result:=m.result;
          exit
       end else
          Ctrl.Handler(m);
    end else begin
       Ctrl:=QWinControl.Create(nil); SetProp(dlg,'Local',integer(Ctrl));
       Ctrl.classname:=cn;
       Ctrl.DoMessage(m);
       if m.result>=1 then begin
          result:=m.result;
          exit
       end else
          Ctrl.Handler(m);
    end;
    result:=m.result;
end;

{ QComponent }
constructor QComponent.Create(AOwner :QComponent);
begin
   inherited;
   fOwner:=AOwner;
   fComponents:=QStringList.Create;
   if fOwner<>nil then
      fOwner.fComponents.AddObject(Self.Name,Self)
end;

destructor QComponent.Destroy;
var
   i :integer;
begin
   if fOwner<>nil then
      fOwner.fComponents.Delete(fOwner.fComponents.IndexOfObject(self));
   for i:=fComponents.Count-1 downto 0 do
       fComponents.Objects[i].Free;
   fComponents.Free;
   inherited;
end;

{ QControl }
constructor QControl.Create(AOwner :QComponent);
begin
   inherited Create(AOwner);
   fControls:=QStringList.Create;
   if fOwner<>nil then
      QControl(fOwner).fControls.AddObject(Self.Name,Self)
end;

destructor QControl.Destroy;
begin
   if fOwner<>nil then
      QControl(fOwner).fControls.Delete(QControl(fOwner).fControls.IndexOfObject(self));
   fControls.Free;
   inherited;
end;

procedure QControl.SetParent(v :QWinControl);
var
    SaveParent :QWinControl;
begin
    SaveParent:=fParent;
    fParent:=v;
    if SaveParent<>nil then begin
       SaveParent.fControls.Delete(SaveParent.fControls.IndexOfObject(self));
       SaveParent.fComponents.Delete(SaveParent.fComponents.IndexOfObject(self));
    end;
    if fParent<>nil then begin
       if self.InheritsFrom(QWinControl) then
          QWinControl(self).fParentWindow:=fParent.fHandle;
       fParent.fControls.AddObject(Name,self);
       fParent.fComponents.AddObject(Name,self);
    end;
    if not isWindow(fHandle) then
       CreateHandle;
end;

procedure QControl.SetText(v:string);
begin
   fText:=v;
   if isWindow(fHandle) then SetWindowText(fHandle,PChar(v));
end;

procedure QControl.SetEnabled(v:boolean);
begin
    fEnabled:=v;
    if isWindow(fHandle) then EnableWindow(fHandle,v);
end;

procedure QControl.SetVisible(v:boolean);
begin
    fVisible:=v;
    if isWindow(fHandle) then
       if v then ShowWindow(fHandle,sw_show) else ShowWindow(fHandle,sw_hide);
end;

procedure QControl.SetLeft(v:integer);
begin
    fleft:=v;
    if isWindow(fHandle) then MoveWindow(fHandle,fleft,ftop,fwidth,fheight,true);
end;

procedure QControl.SetTop(v:integer);
begin
    fTop:=v;
    if isWindow(fHandle) then MoveWindow(fHandle,fleft,ftop,fwidth,fheight,true);
end;

procedure QControl.SetWidth(v:integer);
begin
    fWidth:=v;
    if isWindow(fHandle) then MoveWindow(fHandle,fleft,ftop,fwidth,fheight,true);
end;

procedure QControl.SetHeight(v:integer);
begin
    fHeight:=v;
    if isWindow(fHandle) then MoveWindow(fHandle,fleft,ftop,fwidth,fheight,true);
end;

{ QWinControl }
constructor QWinControl.Create(AOwner :QComponent);
begin
   inherited Create(AOwner);
   fFont:=QFont.Create;
   fEnabled:=true;
   fVisible:=true;
end;

destructor QWinControl.Destroy;
begin
   fFont.Free;
   inherited;
end;

procedure QWinControl.DoMessage(var m:QMessage);
{var
  pif :PPropInfo;
  ocev:QCommandEvent;}
begin
    case m.msg of
    wm_create :begin
               SetWindowLong(fHandle,gwl_id,fControls.Count+1000);
               end;
    wm_command :begin 
                if isWindow(m.lparam) then begin
                   SendMessage(m.lparam,cm_command,loword(m.wparam),hiword(m.wparam));
                end else if hiword(m.wparam)=0 then begin
                    //MENUS
                end else if hiword(m.wparam)=1 then begin
                    //ACCEL
                end else SendMessage(fHandle,cm_command,loword(m.wparam),hiword(m.wparam));
                {ocev:=nil;
                pif :=GetPropInfo(self,'onCommand');
                if pif<>nil then begin
                   @ocev:=FieldAddress('onCommand');
                   if Assigned(ocev) then
                      ocev(Self,loword(m.wparam),hiword(m.wparam),m.lparam);
                end }
                end;
    wm_parentnotify :begin
                     end;
    wm_notify :begin
               end;
    end
end;

procedure QWinControl.CreateHandle;
begin
   CreateControl:=self;
   RegisterClass;
   CreateWindowEx(fExStyle,PChar(ClassName),PChar(fText),fStyle or ws_clipchildren or ws_clipsiblings,fLeft,fTop,fWidth,fHeight,fParentWindow,0,hInstance,nil);
   if isWindow(fHandle) then begin
      SendMessage(fHandle,wm_setfont,fFont.Handle,1);
      if fGrouped then SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) or ws_group);
      if fTabStop then SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) or ws_tabstop);
      EnableWindow(fHandle,fEnabled);
      if fVisible then
         ShowWindow(fHandle,sw_show)
      else
         ShowWindow(fHandle,sw_hide);
   end;
end;

procedure QWinControl.Handler(var m:QMessage);
var
   return :boolean;
   ctrl:QWinControl;
begin
    if m.msg=wm_destroy then begin
       ctrl:=QWinControl(GetProp(m.Dlg,'Local'));
       if Ctrl<>nil then Ctrl.Free
    end;
    wc.cbSize:=sizeof(wc);
    return:=GetClassInfoEx(0,PChar(ClassAncestor),wc);
    if return then
       m.result:=CallWindowProc(wc.lpfnWndProc,m.Dlg,m.msg,m.wparam,m.lparam)
    else
       m.result:=DefWindowProc(m.Dlg,m.msg,m.wparam,m.lparam)
end;

procedure QWinControl.SetFont(v:QFont);
begin
    fFont:=v;
    if isWindow(fHandle) then
       SendMessage(fHandle,wm_setfont,fFont.Handle,1);
end;

procedure QWinControl.SetParentWindow(v:hwnd);
begin
    fParentWindow:=v;
    if not isWindow(fHandle) then
       CreateHandle
    else
       Windows.SetParent(fHandle,v);
    fParent:=QWinControl(GetWindowLong(v,GetClassLong(v,gcl_cbwndextra)-4));
    if fParent<>nil then begin
       fParent.fControls.AddObject(fParent.Name,fParent) ;
       fParent.fComponents.AddObject(fParent.Name,fParent) ;
    end
end;

procedure QWinControl.SetStyle(v:cardinal);
begin
    fStyle:=v;
    if isWindow(fHandle) then begin
       SetWindowLong(fHandle,gwl_style,v);
       SetWindowPos(fHandle,0,0,0,0,0,swp_nosize or swp_nomove or swp_nozorder or swp_noactivate or swp_framechanged);
       Windows.InvalidateRect(fHandle,nil,false);
       Windows.UpdateWindow(fHandle);
    end;
end;

procedure QWinControl.SetExStyle(v:cardinal);
begin
    fExStyle:=v;
    if isWindow(fHandle) then begin
       SetWindowLong(fHandle,gwl_exstyle,v);
       SetWindowPos(fHandle,0,0,0,0,0,swp_nosize or swp_nomove or swp_nozorder or swp_noactivate or swp_framechanged);
       Windows.InvalidateRect(fHandle,nil,false);
       Windows.UpdateWindow(fHandle);
    end;
end;

function QWinControl.RegisterClass:integer;
begin
    wc.cbSize:=sizeof(wc);
    if not(GetClassInfoEx(0,PChar(ClassAncestor),wc)) and not(GetClassInfoEx(hinstance,PChar(ClassAncestor),wc)) then begin
    end;
    wc.cbWndExtra:=wc.cbWndExtra+4;
    wc.lpszClassName:=PChar(ClassName);
    wc.lpfnWndProc:=@GlobalDlgProc;
    wc.hbrBackground:=16;
    wc.hCursor:=LoadCursor(0,idc_arrow);
    wc.hInstance:=hinstance;
    result:=RegisterClassEx(wc);
end;

procedure QWinControl.SetGrouped(v:boolean);
begin
    fGrouped:=v;
    if isWindow(fHandle) then begin
       if v then
          if (GetWindowLong(fHandle,gwl_style) and ws_group)<>ws_group then
              SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) or ws_group);
       if not v then
          if (GetWindowLong(fHandle,gwl_style) and ws_group)=ws_group then
              SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) and not ws_group);
    end;
end;

procedure QWinControl.SetTabStop(v:boolean);
begin
    fTabStop:=v;
    if isWindow(fHandle) then begin
       if v then
          if (GetWindowLong(fHandle,gwl_style) and ws_tabstop)<>ws_tabstop then
              SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) or ws_tabstop);
       if not v then
          if (GetWindowLong(fHandle,gwl_style) and ws_group)=ws_tabstop then
              SetWindowLong(fHandle,gwl_style,GetWindowLong(fHandle,gwl_style) and not ws_tabstop);
    end;
end;

{ QForm }
constructor QForm.Create(AOwner :QComponent);
begin
   inherited Create(AOwner);
   ClassName:='QForm'; 
   fStyle:=ws_overlappedwindow;
   fExStyle:=ws_ex_controlparent;
   fWidth:=400;
   fHeight:=300;
   RegisterClass;
end;

destructor QForm.Destroy;
begin
   if Assigned(App) then
      App.Forms.Delete(App.Forms.IndexOfObject(Self));
   inherited;
end;

procedure QForm.DoMessage(var m:QMessage);
var
  Action:integer;
begin
    inherited;
    case m.msg of
    wm_create: begin
               if Assigned(fCreate) then fCreate(self);
               m.result:=0;
               end;
    wm_close  :begin
               Action:=1;
               if Assigned(fClose) then fClose(self,Action);
               case Action of
               0 :begin
                       m.result:=1;
                       exit
                       end;
               1 :if (GetWindowLong(m.dlg,gwl_exstyle) and ws_ex_appwindow)=ws_ex_appwindow then App.Terminate;
               2 :begin
                       ShowWindow(m.Dlg,sw_minimize);
                       m.result:=1;
                       exit
                       end;
               3 :begin
                       ShowWindow(m.Dlg,sw_maximize);
                       m.result:=1;
                       exit
                       end;
                  end
               end;
    end;
end;

procedure QForm.Handler(var m:QMessage);
begin
    m.result:=DefWindowProc(fHandle,m.msg,m.wparam,m.lparam);
end;

procedure QForm.CreateHandle;
begin
   inherited;
   //do here what you want it
end;

{ QApplication }
constructor QApplication.Create;
begin
    inherited Create(nil);
    fForms:=QStringList.Create;
end;

destructor QApplication.destroy ;
begin
    fForms.Free;
    inherited;
end;

procedure QApplication.DoEvents;
begin
    while PeekMessage(GlobalMsg, 0, 0, 0, PM_REMOVE + PM_NOYIELD) do begin
        TranslateMessage(GlobalMsg);
        DispatchMessage(GlobalMsg);
    end
end;

procedure QApplication.Process(var m:tagMSG);
begin
   inherited;
   if GlobalMsg.message>0 then
      //do something
end;

procedure QApplication.Run;
begin
    while GetMessage(GlobalMsg,0,0,0) do begin
          DispatchMessage(GlobalMsg);
          TranslateMessage(GlobalMsg);
          Process(GlobalMsg);
    end;
end;

procedure QApplication.Terminate;
begin
    ExitProcess(0);
end;

procedure QApplication.Quit;
begin
   PostQuitMessage(0)
end;

procedure QApplication.CreateForm(Ref :QWinControls;var F :Pointer);
begin
    Ref:=QWinControls(GetClass('QMainForm'));
    F := Ref.Create(Self);
    QWinControl(F).Parent:=nil;
end;



initialization
  RegisterClasses([QForm]);

finalization
  unRegisterClasses([QForm]);
  App.Terminate;

end.
